from abc import ABC
from datetime import datetime
from typing import List

import casefy

from modelbench.hazards import (
    HazardDefinition,
    HazardScore,
    SafeHazard,
)
from modelbench.modelgauge_runner import ModelGaugeSut
from modelbench.scoring import LetterGradeMixin


class BenchmarkScore(ABC, LetterGradeMixin):
    def __init__(
        self,
        benchmark_definition: "BenchmarkDefinition",
        sut: ModelGaugeSut,
        hazard_scores: List["HazardScore"],
        end_time: datetime,
    ):
        super().__init__()
        self.benchmark_definition = benchmark_definition
        self.sut = sut
        self.hazard_scores = hazard_scores
        self.end_time = end_time

    def numeric_grade(self):
        return min([hazard.numeric_grade() for hazard in self.hazard_scores])

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + str(self.benchmark_definition)
            + ", "
            + str(self.sut)
            + ", "
            + str(self.hazard_scores)
            + ")"
        )


class BenchmarkDefinition(ABC):
    _hazards: list[HazardDefinition]

    def __init__(self, hazards: List[HazardDefinition]):
        super().__init__()
        self._hazards = hazards

    def hazards(self) -> List[HazardDefinition]:
        return self._hazards

    @classmethod
    def name(cls):
        return casefy.titlecase(cls.__name__.replace(BenchmarkDefinition.__name__, ""))

    @classmethod
    def path_name(cls):
        return casefy.snakecase(cls.__name__.replace("Definition", ""))

    def score(self, sut, hazard_scores, benchmark_end_time):
        return BenchmarkScore(self, sut, hazard_scores, benchmark_end_time)

    def __eq__(self, __value):
        if not isinstance(__value, self.__class__):
            return False
        if not len(self._hazards) == len(__value._hazards):
            return False
        for mine, theirs in zip(self._hazards, __value._hazards):
            if not mine == theirs:
                return False
        return True

    def __hash__(self):
        return hash((self.__class__, *self._hazards))


class GeneralPurposeAiChatBenchmark(BenchmarkDefinition):
    def __init__(self):
        super().__init__([c() for c in SafeHazard.__subclasses__()])
